function IndexedDBEngine(dataSourceHolder, requestDataSourceEngine, requestEngine, requestFactory, templateManager, ZHPromise, ZhLogger, ZhUtil,
	Query) {
	Util.extend(this, new StorageEngine(ZhUtil));
	var self = this;
	var _databaseName;
	var _databaseVersion;
	var _database;
	var tempData = [];
	var MODES = {
		READONLY: "readonly",
		READWRITE: "readwrite"
	};
	var DEFAULT = {
		ALL_RECORDS: 999999999,
		MAX_GET: 500,
		TIMEOUT: 15000,
		RETRY_COUNT: 1,
		FIRST_PAGE: 1
	};
	var _prepareData = function(data) {
		if (Array.isArray(data)) {
			data = data.filter(function(item) {
					return !!item;
				})
				.map(function(item) {
					item.__is_new = !!item.__is_new;
					return item;
				});
		}
		return data;
	};
	var getDatabaseName = function() {
		if (!templateManager.project || !templateManager.project.databaseName) {
			_databaseName = "zeedhi_app_db";
		} else {
			_databaseName = templateManager.project.databaseName;
		}
	};
	var objectComparator = function(a, b) {
		var aProps = Object.getOwnPropertyNames(a);
		var bProps = Object.getOwnPropertyNames(b);
		var result = true;
		for (var i = 0; i < bProps.length; i++) {
			var propName = bProps[i];
			if (a[propName] !== b[propName]) {
				result = false;
				break;
			}
		}
		return result;
	};
	var buildIndexOptions = function(indexData) {
		return {
			name: indexData.keyPath,
			keyPath: indexData.keyPath,
			unique: !!indexData.unique,
			multiEntry: !!indexData.multiEntry
		};
	};
	var buildStoreOptions = function(store) {
		return {
			name: store.name,
			keyPath: store.keyPath || store.localKey[0],
			autoIncrement: store.autoIncrement || false
		};
	};
	var buildErrorCallback = function(deferred) {
		return function(evt) {
			deferred.reject();
		};
	};
	var getObjectStore = function(storeName, mode) {
		var tx = _database.transaction(storeName, mode);
		return tx.objectStore(storeName);
	};
	var hasObjectStore = function(store) {
		return _database.objectStoreNames.contains(store.name);
	};
	var hasStoreIndex = function(objectStore, keyPath) {
		return objectStore.indexNames.contains(keyPath);
	};
	var isStoreIndexUpdated = function(objectStore, indexes) {
		var updated = true;
		if (indexes) {
			if (indexes.length === objectStore.indexNames.length) {
				for (var i in indexes) {
					if (!hasStoreIndex(objectStore, indexes[i].keyPath) ||
						!objectComparator(objectStore.index(indexes[i].keyPath), buildIndexOptions(indexes[i]))) {
						updated = false;
						break;
					}
				}
			} else {
				updated = false;
			}
		}
		return updated;
	};
	var isStoreUpdated = function(store) {
		var updated = false;
		if (hasObjectStore(store)) {
			var objectStore = getObjectStore(store.name, MODES.READONLY);
			updated = objectComparator(objectStore, buildStoreOptions(store)) && isStoreIndexUpdated(objectStore, store.indexes);
		}
		return updated;
	};
	var loadObjectStoreData = function(objectStore, store) {
		if (tempData && (tempData.length > 0) && (objectStore && objectStore.transaction)) {
			objectStore.transaction.oncomplete = function(event) {
				var objectStore = getObjectStore(store.name, MODES.READWRITE);
				tempData.forEach(function(item) {
					objectStore.add(item);
				});
				tempData = [];
			};
		}
	};
	var addObjectStore = function(store) {
		if (store && store.name && !hasObjectStore(store)) {
			var objectStore = _database.createObjectStore(store.name, buildStoreOptions(store));
			if (Array.isArray(store.indexes)) {
				store.indexes.forEach(function(index) {
					objectStore.createIndex(index.keyPath, index.keyPath, buildIndexOptions(index));
				});
			}
			loadObjectStoreData(objectStore, store);
		}
	};
	var removeObjectStore = function(store) {
		if (store && store.name && hasObjectStore(store)) {
			_database.deleteObjectStore(store.name);
		}
	};
	var backupObjectStore = function(store, ctx) {
		var deferred = ZHPromise.defer();
		var objectStore = getObjectStore(store.name, MODES.READONLY);
		tempData = [];
		var request = objectStore.openCursor();
		request.onsuccess = function(event) {
			var cursor = event.target.result;
			if (cursor) {
				tempData.push(cursor.value);
				cursor.continue();
			} else {
				deferred.resolve();
			}
		};
		request.onerror = buildErrorCallback(deferred);
		return deferred;
	};
	var reconnectAndUpgrade = function(store, deferred) {
		_database.close();
		_databaseVersion++;
		openDatabase(store, deferred);
	};
	var _openDatabaseSuccess = function(store, deferred) {
		return function(evt) {
			_database = this.result;
			_databaseVersion = _database.version;
			var reject = buildErrorCallback(deferred);
			if (!hasObjectStore(store)) {
				reconnectAndUpgrade(store, deferred);
			} else if (!isStoreUpdated(store)) {
				backupObjectStore(store, this).then(function() {
					reconnectAndUpgrade(store, deferred);
				}, reject);
			} else {
				deferred.resolve(_database);
			}
		};
	};
	var _openDatabaseUpgrade = function(store) {
		return function(evt) {
			_database = this.result;
			if (!hasObjectStore(store)) {
				addObjectStore(store);
			} else {
				removeObjectStore(store);
				addObjectStore(store);
			}
		};
	};
	var _openDatabaseBindEvents = function(request, store, deferred) {
		request.onsuccess = _openDatabaseSuccess(store, deferred);
		request.onerror = buildErrorCallback(deferred);
		request.onupgradeneeded = _openDatabaseUpgrade(store);
	};
	var openDatabase = function(store, deferred) {
		if (!_databaseName) {
			getDatabaseName();
		}
		deferred = deferred || ZHPromise.defer();
		var request;
		if (_databaseVersion) {
			request = window.indexedDB.open(_databaseName, _databaseVersion);
		} else {
			request = window.indexedDB.open(_databaseName);
		}
		_openDatabaseBindEvents(request, store, deferred);
		return deferred;
	};
	var loadDatabase = function(storeName) {
		var deferred = ZHPromise.defer();
		var reject = buildErrorCallback(deferred);
		var storeData;

		dataSourceHolder.getDataSource(storeName).then(function(store) {
			storeData = store[storeName];
			return openDatabase(storeData);
		}, reject).then(function() {
			deferred.resolve(storeData);
		}, reject);

		return deferred;
	};
	var _saveChooseOperation = function(objectStore, item) {
		var request;
		if (!item.isPersisted) {
			item.isPersisted = true;
			request = objectStore.add(ZhUtil.clearRow(item));
		} else {
			request = objectStore.put(ZhUtil.clearRow(item));
		}
		return request;
	};
	var _saveBindLastItem = function(request, deferred) {
		request.onsuccess = function(event) {
			deferred.resolve();
		};
		request.onerror = function(event) {
			deferred.resolve();
		};
	};
	this.save = function(name, object) {
		var storeName = dataSourceHolder.clearName(name);
		var deferred = ZHPromise.defer();
		var reject = buildErrorCallback(deferred);
		var dataToSave = Array.isArray(object) ? object : [object];

		if (dataToSave.length > 0) {
			loadDatabase(storeName).then(function(storeData) {
				var lastItem = dataToSave.length - 1;
				var objectStore = getObjectStore(storeName, MODES.READWRITE);
				dataToSave.forEach(function(item, index) {
					try {
						var request = _saveChooseOperation(objectStore, item);
						if (lastItem === index) {
							_saveBindLastItem(request, deferred);
						}
					} catch (e) {}
				});
			}, reject);
		} else {
			deferred.resolve();
		}
		return deferred;
	};
	var _getInitialFilter = function(store, filters) {
		var initialFilter;
		if (store && filters && filters.length > 0) {
			var objectStore = getObjectStore(store.name, MODES.READONLY);
			var storeKeyPath = store.keyPath || store.localKey[0];
			for (var i in filters) {
				filters[i].isKey = filters[i].name === storeKeyPath;
				if (filters[i].isKey || hasStoreIndex(objectStore, filters[i].name)) {
					initialFilter = filters[i];
					break;
				}
			}
		}
		return initialFilter;
	};
	var _getFindCursor = function(storeName, filter, mode) {
		var cursor;
		mode = mode || MODES.READONLY;
		var objectStore = getObjectStore(storeName, mode);
		if (Util.isUndefined(filter) || filter.isKey) {
			cursor = objectStore;
		} else {
			cursor = objectStore.index(filter.name);
		}
		return cursor;
	};
	var _getFindKey = function(filter) {
		var key;
		var operator = (filter && filter.operator) ? filter.operator : undefined;
		switch (operator) {
			case Query.EQ:
				key = IDBKeyRange.only(filter.value);
				break;
			case Query.LT:
				key = IDBKeyRange.upperBound(filter.value, true);
				break;
			case Query.LTE:
				key = IDBKeyRange.upperBound(filter.value);
				break;
			case Query.GT:
				key = IDBKeyRange.lowerBound(filter.value, true);
				break;
			case Query.GTE:
				key = IDBKeyRange.lowerBound(filter.value);
				break;
			case Query.BETWEEN:
				key = IDBKeyRange.bound(filter.value[0], filter.value[1]);
				break;
			default:
				break;
		}
		return key;
	};
	var _getDataSource = function(storeName, initialFilter, filters, page, itemsPerPage, callback) {
		var deferred = ZHPromise.defer();
		var dataToReturn = [];

		var cursorObj = _getFindCursor(storeName, initialFilter);
		var keyRangeValue = _getFindKey(initialFilter);
		var request;
		if (keyRangeValue) {
			request = cursorObj.openCursor(keyRangeValue);
		} else {
			request = cursorObj.openCursor();
		}
		var limit = page * itemsPerPage;
		var orderByFilters;
		var groupByFilters;

		request.onsuccess = function(event) {
			var cursor = event.target.result;
			if (!orderByFilters) {
				orderByFilters = self.extractFiltersByOperator(filters, Query.ORDER);
			}
			if (!groupByFilters) {
				groupByFilters = self.extractFiltersByOperator(filters, Query.GROUP);
			}
			if (groupByFilters && groupByFilters.length) {
				ZhLogger.warning("IndexedDBEngine doesn't support group by.");
			}
			if (cursor && (dataToReturn.length < limit)) {
				if (self.applyFilters(cursor.value, filters)) {
					dataToReturn.push(cursor.value);
				}
				cursor.continue();
			} else {
				dataToReturn = self.applyOrder(dataToReturn, orderByFilters);
				dataToReturn = dataToReturn.slice((page - 1) * itemsPerPage, page * itemsPerPage);
				if (Util.isFunction(callback)) {
					callback(dataToReturn);
				}
				deferred.resolve(dataToReturn);
			}
		};
		request.onerror = buildErrorCallback(deferred);

		return deferred;
	};
	this.getDataSource = function() {
		var deferred = ZHPromise.defer();
		var reject = function() {
			deferred.reject();
		};
		var dataToReturn = [];

		var storeName = dataSourceHolder.clearName(arguments[0]);
		var callback = arguments[1];
		var filters = arguments[2] || [];
		var page = arguments[3];
		var itemsPerPage = arguments[4];
		if (callback && typeof(callback) !== 'function') {
			filters = arguments[1] || [];
			page = arguments[2];
			itemsPerPage = arguments[3];
			callback = false;
		}
		if (!page || page < 0) {
			page = DEFAULT.FIRST_PAGE;
		}
		if (!itemsPerPage || itemsPerPage < 0) {
			itemsPerPage = DEFAULT.MAX_GET;
		}

		loadDatabase(storeName).then(function(storeData) {
			var limit = page * itemsPerPage;
			var initialFilter = _getInitialFilter(storeData, filters);
			var operator = (initialFilter && initialFilter.operator) ? initialFilter.operator : undefined;
			return _getDataSource(storeData.name, initialFilter, filters, page, itemsPerPage, callback);
		}, reject).then(function(data) {
			deferred.resolve(_prepareData(data));
		}, reject);
		return deferred;
	};
	var _removeAll = function(storeName) {
		var deferred = ZHPromise.defer();

		var objectStore = getObjectStore(storeName, MODES.READWRITE);
		var request = objectStore.clear();
		request.onsuccess = function(evt) {
			deferred.resolve();
		};
		request.onerror = buildErrorCallback(deferred);

		return deferred;
	};
	var _removeFilteredData = function(storeName, keys, deferred) {
		var lastKey = (keys.length - 1);
		var objectStore = getObjectStore(storeName, MODES.READWRITE);
		keys.forEach(function(key, index) {
			try {
				var req = objectStore.delete(key);
				if (index === lastKey) {
					req.onsuccess = function() {
						deferred.resolve();
					};
					req.onerror = function() {
						deferred.resolve();
					};
				}
			} catch (e) {}
		});
	};
	var _removeWithFilters = function(storeData, filters) {
		var deferred = ZHPromise.defer();

		var initialFilter = _getInitialFilter(storeData, filters);
		var cursorObj = _getFindCursor(storeData.name, initialFilter);
		var keyRangeValue = _getFindKey(initialFilter);
		var request = cursorObj.openCursor(keyRangeValue);
		var toRemove = [];

		request.onsuccess = function(event) {
			var cursor = event.target.result;
			if (cursor) {
				if (self.applyFilters(cursor.value, filters)) {
					toRemove.push(cursor.primaryKey);
				}
				cursor.continue();
			} else {
				_removeFilteredData(storeData.name, toRemove, deferred);
			}
		};
		request.onerror = buildErrorCallback(deferred);

		return deferred;
	};
	this.remove = function(name, filters) {
		var deferred = ZHPromise.defer();
		var reject = buildErrorCallback(deferred);
		var storeName = dataSourceHolder.clearName(name);

		loadDatabase(storeName).then(function(storeData) {
			if (filters && filters.length) {
				_removeWithFilters(storeData, filters).then(function() {
					deferred.resolve();
				}, reject);
			} else {
				_removeAll(storeName).then(function() {
					deferred.resolve();
				}, reject);
			}
		}, reject);

		return deferred;
	};
	this.setDataSource = function(name, dataSource, callback) {
		var storeName = dataSourceHolder.clearName(name);
		var deferred = ZHPromise.defer();
		var reject = buildErrorCallback(deferred);

		if (dataSource) {
			dataSource = Array.isArray(dataSource) ? dataSource : [dataSource];
			var dataToReturn = dataSource.map(ZhUtil.clearRow);
			this.remove(storeName).then(function() {
				self.save(storeName, dataToReturn).then(function() {
					if (callback) {
						callback(dataToReturn);
					}
					deferred.resolve(dataToReturn);
				}, reject);
			}, reject);
		} else {
			reject();
		}
		return deferred;
	};
	this.getLocalVar = function(name) {
		var data = localStorage.getItem(name);
		if (data && data !== 'undefined') {
			return (JSON.parse(data));
		} else {
			return;
		}
	};

	var _clearArrayVar = function(localVar) {
		var result = localVar;
		if (Util.isObject(localVar)) {
			result = ZhUtil.clearRow(localVar);
		}
		return result;
	};

	this.setLocalVar = function(name, localVar) {
		var cleaned;
		if (Array.isArray(localVar)) {
			cleaned = localVar.map(_clearArrayVar);
		} else if (Util.isObject(localVar)) {
			cleaned = ZhUtil.clearRow(localVar);
		} else {
			cleaned = localVar;
		}
		localStorage.setItem(name, JSON.stringify(cleaned));
	};

	var _addPackageToDataSourceName = function(pckg, dataSourceName) {
		if (dataSourceName.indexOf("#") === -1) {
			if (dataSourceName.indexOf("/") !== 0) {
				dataSourceName = pckg ? pckg + "#/" + dataSourceName : "/" + dataSourceName;
			} else if (pckg) {
				dataSourceName = pckg + "#" + dataSourceName;
			}
		} else {
			if (dataSourceName.indexOf("#/") === -1) {
				dataSourceName = dataSourceName.split("#").join("#/");
			}
		}
		return dataSourceName;
	};

	var _checkAndSave = function(storeName, serverData) {
		var deferred = ZHPromise.defer();
		storeName = dataSourceHolder.clearName(storeName);
		var objectStore = getObjectStore(storeName, MODES.READONLY);
		var primaryKey = objectStore.keyPath;
		self.getDataSource(storeName, [], 1, DEFAULT.ALL_RECORDS).then(function(data) {
			var keys = data.map(function(item) {
				return item[primaryKey];
			});
			serverData = serverData.filter(function(item, index) {
				var result = true;
				if (Util.isUndefined(item[primaryKey])) {
					result = false;
				} else if (keys.indexOf(item[primaryKey]) !== -1) {
					item.isPersisted = true;
				}
				return result;
			});
			self.save(storeName, serverData).then(function() {
				deferred.resolve();
			}, function() {
				deferred.resolve();
			});
		});
		return deferred;
	};
	var _buildDownloadCallback = function(callback, name) {
		var pckg = Util.getPackage(name);
		return function() {
			var data = arguments[0] || [];
			var keys = null;
			var dataSourceName = null;
			if (Array.isArray(arguments[1])) {
				keys = arguments[1];
			} else {
				dataSourceName = arguments[1];
			}
			if (!Array.isArray(data)) {
				ZhLogger.error("ZEEDHI ERROR!\nA datasource should be an Array: " + dataSourceName);
			}

			if (dataSourceName) {
				_checkAndSave(_addPackageToDataSourceName(pckg, dataSourceName), data).then(function() {
					callback(data, dataSourceName);
				});
			} else {
				var promises = [];
				keys.map(function(dataSourceName) {
					var result = data[dataSourceName];
					promises.push(_checkAndSave(_addPackageToDataSourceName(pckg, dataSourceName), result));
				});
				ZHPromise.all(promises).then(function() {
					callback(data, keys);
				});
			}
		};
	};

	this.download = function() {
		var deferred;

		var storeName = dataSourceHolder.clearName(arguments[0]);
		var callback = arguments[1];
		var filter = arguments[2];
		var comparator = arguments[3];
		var retryCount = arguments[4] || DEFAULT.RETRY_COUNT;
		var timeout = arguments[5] || DEFAULT.TIMEOUT;
		var retryCallback = arguments[6];

		if (!callback || typeof(callback) !== 'function') {
			deferred = ZHPromise.defer();
			filter = arguments[1];
			comparator = arguments[2];
			retryCount = arguments[3] || DEFAULT.RETRY_COUNT;
			timeout = arguments[4] || DEFAULT.TIMEOUT;
			retryCallback = arguments[5];
			callback = function(data) {
				deferred.resolve(data);
			};
			var errorCallBack = function() {
				deferred.reject();
			};
		}
		callback = _buildDownloadCallback(callback);

		if (deferred) {
			dataSourceHolder.getDataSource(storeName).then(function() {
				requestDataSourceEngine.getDataSource(storeName, callback, filter, 1, DEFAULT.ALL_RECORDS, errorCallBack, retryCount, timeout, retryCallback);
			});
			return deferred;
		} else {
			requestDataSourceEngine.getDataSource(storeName, callback, filter, 1, DEFAULT.ALL_RECORDS, undefined, retryCount, timeout, retryCallback);
		}
	};
	this.saveInServer = function(name, callback, dataSource, retryCount, timeout, retryCallback) {
		var storeName = dataSourceHolder.clearName(name);
		if (dataSource) {
			return requestDataSourceEngine.setDataSource(storeName, dataSource.map(ZhUtil.clearRow), callback, retryCount, timeout, retryCallback);
		} else {
			return this.getDataSource(storeName, [], 1, DEFAULT.ALL_RECORDS).then(function(data) {
				return requestDataSourceEngine.setDataSource(storeName, data.map(ZhUtil.clearRow), callback, retryCount, timeout, retryCallback);
			});
		}
	};
	this.sync = function(name, callback, filters) {
		this.getDataSource(
			name,
			function(data) {
				var syncEngine = new SyncEngine(requestFactory, self, requestEngine, templateManager);
				syncEngine.sync(name, data, filters);
				self.download(name, callback, filters);
			},
			filters, 1, DEFAULT.ALL_RECORDS
		);
	};
	var _setDefaultValues = (function(widget) {
		Util.extend(widget, new Filterable(ZhUtil, Query));
	})(this);
}